source('~/s2dverification/R/Utils.R')
library(ncdf4)
library(abind)

.CheckSelectors <- function(selectors, var = NULL, return_indices = TRUE) {
  if (return_indices) {
    if (is.list(selectors)) {
      if (length(selectors) != 2) {
        stop("'selectors' provided in a wrong format.")
      }
      for (i in 1:length(selectors)) {
        if (is.null(var)) {
          stop("No selector values provided in 'var'.")
        } else if (is.na(selectors[[i]])) {
          if (i == 1) {
            selectors[[i]] <- 1
          } else {
            selectors[[i]] <- length(var)
          }
        } else if (is.character(selectors[[i]])) {
          candidate <- which(var == selectors[[i]])
          if (length(candidate) > 0) {
            selectors[[i]] <- candidate[1]
          } else {
            stop("Selector value not found in 'var'.")
          }
        }
      }
      selectors[[1]]:selectors[[2]]
    } else if (is.numeric(selectors)) {
      if (length(selectors) > 0) {
        selectors
      } else {
        stop("No selectors provided in 'selectors'.")
      }
    } else if (is.character(selectors)) {
      if (is.null(var)) {
        stop("No selector values provided in 'var'.")
      } else {
        if ((length(selectors) == 1) && 
            (selectors %in% c('all', 'first', 'last'))) {
          if (selectors == 'all') {
            1:length(var)
          } else if (selectors == 'first') {
            1
          } else {
            length(var)
          }
        } else {
          candidates <- which(var == selectors)
          if (length(candidates) == 0) {
            stop("Selectors do not match values in 'var'.")
          } else if (length(candidates) != length(selectors)) {
            stop("Some selectors do not match values in 'var'.")
          }
          candidates
        }
      }
    }
  } else {
    if (is.list(selectors)) {
    } else if (is.numeric(selectors)) {
      if (length(selectors) > 0) {
        if (!is.null(var)) {
          var[selectors]
        }
      } else {
        stop("No selectors provided.")
      }
    } else if (is.character(selectors)) {
      if ((length(selectors) == 1) && 
          (selectors %in% c('all', 'first', 'last'))) {
        if (selectors == 'all') {
          var
        } else if (selectors == 'first') {
          head(var, 1)
        } else {
          tail(var, 1)
        }
      } else {
        selectors
      }
    }
  }
}

## TODO: Somehow reorder lons and lats
.NcOpener <- function(file_name) {
  result <- NULL
  
  try({
    result <- nc_open(file_name)
  })
  
  result
}

.NcVarReader <- function(file_name = NULL, file_object = NULL, var_name = NULL) {
  if (is.null(file_object)) {
    if (is.null(file_name)) {
      stop("Either 'file_name' or 'file_object' must be provided.")
    }
    file_object <- .NcOpener(file_name)
  } else {
    file_name <- file_object$filename
  }
  if (is.null(var_name)) {
    if (length(fnc$var) == 0) {
      stop(paste0("No variables found in ", file_name, "."))
    } else if (length(fnc$var) > 1) {
      stop("A 'var_name' must be provided if the file contains more than one variables.")
    } else {
      var_name <- names(fnc$var)[1]
    }
  }

  result <- NULL
  try({
    result <- ncvar_get(file_object, var_name)
    units <- file_object$var[[var_name]]$units
    if (units %in% c('seconds', 'minutes', 'hours', 'days', 'weeks', 'months', 'years')) {
      if (units == 'seconds') {
        units <- 'secs'
      } else if (units == 'minutes') {
        units <- 'mins'
      }
      result <- paste(result, units)
    } else if (grep(' since ', units)) {
      parts <- strsplit(units, ' since ')[[1]]
      units <- parts[1]
      if (units == 'seconds') {
        units <- 'secs'
      } else if (units == 'minutes') {
        units <- 'mins'
      }
      result <- seq(as.POSIXct(parts[2]), length = max(result, na.rm = TRUE), by = units)
    }
  })

  result
}

.NcDataReader <- function(file_name, file_object, indices, vars) {#, file_dims) {
  if (is.null(variable_name)) {
  }
  ## TODO: check performance of wrapping opener functions
  ncvar_get(file_object, variable_name)
}

.NcCloser <- function(file_pointer) {
  result <- NULL

  try({
    result <- nc_close(file_pointer)
  })

  invisible(result)
}

.Load <- function(..., pattern_dims = NULL, 
                  inner_dims_across_files = NULL,   # ex list(ftime = 'chunk')
                  return_vars = NULL,               # ex list('ftime_var', 'lat_var', lon = c('lon', 'longitude'), ftime = 'extra_var'). 
                                                    # Allow only one dimension per variable? No .
                                                    # Here return the names without the _var.
                                                    # ex list(ftime_var = NULL, lat_var = NULL, lon = NULL, date = c('sdate', 'ftime'))
                                                    #                                                       date = list(var = 'time',
                                                    #                                                                   dims = c('ftime'),
                                                    #                                                                   for_each = c('sdate', 'ftime'))
                                                    # by default should take NULL.
                  selector_checker  = .CheckSelectors,
                  file_opener = .NcOpener,
                  file_var_reader = .NcVarReader,
                  file_data_reader = .NcDataReader,
                  file_closer = .NcCloser) {
                  #, config_file = NULL
                  #, variable_dim = , 
                  #file_dimnames = , 
                  #) { This one is needed if user wants to get values for a variable that is not defined as a *_var.
  dim_params <- list(...)
  # Split *_var parameters
  var_params_ind <- grep('_var$', names(dim_params))
  var_params <- dim_params[var_params_ind]
  # Check all *_var are NULL or vectors of character strings, and 
  # that they all have a matching dimension param.
  i <- 1
  for (var_param in var_params) {
    if (!is.null(var_param)) {
      if (!is.character(var_param)) {
        stop("All '*_var' parameters must be vectors of character strings or NULL.")
      } else if (!any(grepl(paste0('^', strsplit(names(var_params)[i], 
                                                 '_var$')[[1]][1], '$'),
                            names(dim_params)))) {
        stop(paste0("All '*_var' parameters must be associated to a dimension parameter. Found parameter '", 
                    names(var_params)[i], "' but no parameter '", 
                    strsplit(names(var_params)[i], '_var$')[[1]][1], "'."))
      }
    }
    i <- i + 1
  }
  # Split *_checker parameters
  check_params_ind <- grep('_checker$', names(dim_params))
  check_params <- dim_params[check_params_ind]
  # Check all *_checker are NULL or functions, and that they all have 
  # a matching dimension param.
  i <- 1
  for (check_param in check_params) {
    if (!is.null(check_param)) {
      if (!is.function(check_param)) {
        stop("All '*_checker' parameters must be functions or NULL.")
      } else if (!any(grepl(paste0('^', strsplit(names(check_params)[i], 
                                                 '_checker$')[[1]][1], '$'),
                            names(dim_params)))) {
        stop(paste0("All '*_checker' parameters must be associated to a dimension parameter. Found parameter '", 
                    names(check_params)[i], "' but no parameter '", 
                    strsplit(names(check_params)[i], '_checker$')[[1]][1], "'."))
      }
    }
    i <- i + 1
  }
  # Leave alone the dimension parameters
  if (length(c(var_params_ind, check_params_ind)) > 0) {
    dim_params <- dim_params[-c(var_params_ind, check_params_ind)]
  }
  dim_names <- names(dim_params)
  # Make the keys of 'var_params' and 'check_params' to be the name of 
  # the corresponding dimension.
  names(var_params) <- gsub('_var$', '', names(var_params))
  names(check_params) <- gsub('_checker$', '', names(check_params))

  # Check pattern_dims
  if (is.null(pattern_dims)) {
    print(paste0("Parameter 'pattern_dims' not specified. Taking the first dimension, '", 
                 dim_names[1], "' as 'pattern_dims'."))
    pattern_dims <- dim_names[1]
  } else if (is.character(pattern_dims) && (length(pattern_dims) > 0)) {
    pattern_dims <- unique(pattern_dims)
  } else {
    stop("Parameter 'pattern_dims' must be a vector of character strings.")
  }
  if (any(names(var_params) %in% pattern_dims)) {
    stop("'*_var' parameters specified for pattern dimensions. Remove or fix them.")
  }

  ## TODO: varmin and varmax
  # Find the pattern dimension with the pattern specifications
  found_pattern_dim <- NULL
  for (pattern_dim in pattern_dims) {
    # Check all specifications in pattern_dim are valid
    dat <- datasets <- dim_params[[pattern_dim]]
    if (is.null(dat) || !(is.character(dat) && all(nchar(dat) > 0)) && !is.list(dat)) {
      stop(paste0("Parameter '", pattern_dim, 
                  "' must be a list of lists with pattern specifications or a vector of character strings."))
    }
    if (!is.null(check_params[[pattern_dim]])) {
      print(paste0("A checker for the selectors of '", pattern_dim, 
                   "' has been specified, but it is a pattern dimension and the checker will be ignored."))
    }
    if (any(sapply(dat, is.list))) {
      if (is.null(found_pattern_dim)) {
        found_pattern_dim <- pattern_dim
      } else {
        stop("Found more than one pattern dim with pattern specifications (list of lists). One and only one pattern dim must contain pattern specifications.")
      }
    }
  }
  if (is.null(found_pattern_dim)) {
    print(paste0("Taking the first pattern dim, '", pattern_dims[1], "', as dimension with pattern specifications (to be fetched in configuration file)."))
    found_pattern_dim <- pattern_dims[1]
  }
  dat <- datasets <- dim_params[[found_pattern_dim]]
  dat_info_names <- c('name', 'path')#, 'nc_var_name', 'suffix', 'var_min', 'var_max', 'dimnames')
  dat_to_fetch <- c()
  dat_names <- c()
  if (!is.list(dat)) {
    dat <- lapply(dat, function (x) list(name = x))
  }
  for (i in 1:length(dat)) {
    if (is.character(dat[[i]]) && length(dat[[i]]) == 1 && nchar(dat[[i]]) > 0) {
      dat[[i]] <- list(name = dat[[i]])
    } else if (!is.list(dat[[i]])) {
      stop(paste0("Parameter '", pattern_dim, 
                  "' is incorrect. It must be a list of lists or character strings."))
    }
    #if (!(all(names(dat[[i]]) %in% dat_info_names))) {
    #  stop("Error: parameter 'dat' is incorrect. There are unrecognized components in the information of some of the dateriments. Check 'dat' in ?Load for details.")
    #}
    if (!('name' %in% names(dat[[i]]))) {
      dat[[i]][['name']] <- paste0('dat', i)
      if (!('path' %in% names(dat[[i]]))) {
        stop(paste0("Parameter '", found_pattern_dim, 
                    "' is incorrect. A 'path' should be provided for each dataset if no 'name' is provided."))
      }
    } else if (!('path' %in% names(dat[[i]]))) {
      dat_to_fetch <- c(dat_to_fetch, i)
    }
    #if ('path' %in% names(dat[[i]])) {
    #  if (!('nc_var_name' %in% names(dat[[i]]))) {
    #    dat[[i]][['nc_var_name']] <- '$var_name$'
    #  }
    #  if (!('suffix' %in% names(dat[[i]]))) {
    #    dat[[i]][['suffix']] <- ''
    #  }
    #  if (!('var_min' %in% names(dat[[i]]))) {
    #    dat[[i]][['var_min']] <- ''
    #  }
    #  if (!('var_max' %in% names(dat[[i]]))) {
    #    dat[[i]][['var_max']] <- ''
    #  }
    #}
    dat_names <- c(dat_names, dat[[i]][['name']])
  }
  if ((length(dat_to_fetch) > 0) && (length(dat_to_fetch) < length(dat))) {
    print("'path' has been provided for some datasets. Any information in the configuration file related to these will be ignored.")
  }

  ## TODO: Check opener, reader, etc... functions are specified

  ## TODO: check config_file
  ## TODO: Get rid of this default
  #if (is.null(config_file)) {
  #  config_file <- system.file("config", "BSC.conf", package = "s2dverification")
  #} else if (!is.character(config_file) || !(nchar(config_file) > 0)) {
  #  stop("Error: parameter 'config_file' must be a character string with the path to a .Load() configuration file, if provided.")
  #}

  ## If not all data has been provided for any of the datasets in the 
  ## pattern_dim, configuration file is read.
  #if (length(dat_to_fetch) > 0) {
  #  print(paste0("Some 'path's not explicitly provided in '", pattern_dim, 
  #               "', so will now proceed to open the configuration file.")
  #  data_info <- ConfigFileOpen(config_file, silent, TRUE)
  #
  #  # Check that the var, exp and obs parameters are right and keep the entries
  #  # that match for each dataset.
  #  # Afterwards, the matching entries are applied sequentially (as specified 
  #  # in ?ConfigFileOpen) and the replace_values are applied to the result.
  #  # Finally a path pattern for each dataset is provided.
  #  matches <- ConfigApplyMatchingEntries(data_info, var, sapply(exp[exps_to_fetch], '[[', 'name'),
  #               sapply(obs[obs_to_fetch], '[[', 'name'), show_entries = FALSE, show_result = FALSE)
  #  # 'replace_values' is a named list that associates a variable name to an 
  #  # associated value. Initially it is filled with variables and values parsed
  #  # from the configuration file, but we can add or modify some values during
  #  # the execution to choose for example which start date we want to load.
  #  # When '.ConfigReplaceVariablesInString' is called, all the variable accesses 
  #  # ($VARIABLE_NAME$) that appear in the string given as parameter are 
  #  # replaced by the associated value in 'replace_values'.
  #  replace_values <- data_info$definitions
  #  if (!is.null(exp) && length(exps_to_fetch) > 0) {
  #    counter <- 1
  #    exp[exps_to_fetch] <- lapply(matches$exp_info,
  #      function (x) {
  #        x[names(exp[[exps_to_fetch[counter]]])] <- exp[[exps_to_fetch[counter]]]
  #        x[['path']] <- paste0(x[['main_path']], x[['file_path']])
  #        counter <<- counter + 1
  #        x
  #      })
  #  }
  #  if (!is.null(obs) && length(obs_to_fetch) > 0) {
  #    counter <- 1
  #    obs[obs_to_fetch] <- lapply(matches$obs_info,
  #      function (x) {
  #        x[names(obs[[obs_to_fetch[counter]]])] <- obs[[obs_to_fetch[counter]]]
  #        x[['path']] <- paste0(x[['main_path']], x[['file_path']])
  #        counter <<- counter + 1
  #        x
  #      })
  #  }
  #  if (!silent) {
  #    .message("All pairs (var, exp) and (var, obs) have matching entries.")
  #  }
  #} else {
  #  replace_values <- list(DEFAULT_NC_VAR_NAME = '$VAR_NAME$',
  #                         DEFAULT_VAR_MIN = '',
  #                         DEFAULT_VAR_MAX = '',
  #                         DEFAULT_SUFFIX = '',
  #                         DEFAULT_DIM_NAME_LONGITUDES = 'longitude',
  #                         DEFAULT_DIM_NAME_LATITUDES = 'latitude',
  #                         DEFAULT_DIM_NAME_MEMBERS = 'ensemble')
  #}
  ## We take the dimnames that haven't been explicitly specified from the 
  ## configuration file.
  ## If the configuration file wasn't opened, we take the default values from
  ## the dictionary 'replace_values'.
  #dimnames <- list(lon = ifelse(is.null(dimnames[["lon"]]),
  #                              replace_values[["DEFAULT_DIM_NAME_LONGITUDES"]],
  #                              dimnames[['lon']]),
  #                 lat = ifelse(is.null(dimnames[["lat"]]),
  #                              replace_values[["DEFAULT_DIM_NAME_LATITUDES"]],
  #                              dimnames[['lat']]),
  #                 member = ifelse(is.null(dimnames[["member"]]),
  #                                 replace_values[["DEFAULT_DIM_NAME_MEMBERS"]],
  #                                 dimnames[['member']]))
  #mandatory_defaults <- c('DEFAULT_EXP_MAIN_PATH', 'DEFAULT_EXP_FILE_PATH',
  #                        'DEFAULT_OBS_MAIN_PATH', 'DEFAULT_OBS_FILE_PATH',
  #                        'DEFAULT_NC_VAR_NAME', 'DEFAULT_SUFFIX',
  #                        'DEFAULT_VAR_MIN', 'DEFAULT_VAR_MAX',
  #                        'DEFAULT_DIM_NAME_LONGITUDES',
  #                        'DEFAULT_DIM_NAME_LATITUDES',
  #                        'DEFAULT_DIM_NAME_MEMBERS')
  #extra_vars_with_default_ind <- (1:length(replace_values))[grep('^DEFAULT_', names(replace_values))]
  #extra_vars_with_default_ind <- extra_vars_with_default_ind[
  #                                 grep(paste0(paste0('^', mandatory_defaults),
  #                                             collapse = '|'),
  #                                      names(replace_values)[extra_vars_with_default_ind],
  #                                      invert = TRUE)
  #                               ]
  #extra_vars_with_default <- gsub('^DEFAULT_', '',
  #                                names(replace_values)[extra_vars_with_default_ind])
  #if (!is.null(exp)) {
  #  exp <- lapply(exp, function (x) {
  #    if (!('dimnames' %in% names(x))) {
  #      x[['dimnames']] <- dimnames
  #    } else {
  #      dimnames2 <- dimnames
  #      dimnames2[names(x[['dimnames']])] <- x[['dimnames']]
  #      x[['dimnames']] <- dimnames2
  #    }
  #    i <- 1
  #    while (i <= length(extra_vars_with_default)) {
  #      if (!(extra_vars_with_default[i] %in% names(x))) {
  #        x[[extra_vars_with_default[i]]] <- replace_values[[extra_vars_with_default_ind[i]]]
  #      }
  #      i <- i + 1
  #    }
  #    x
  #  })
  #}
  #if (!is.null(obs)) {
  #  obs <- lapply(obs, function (x) {
  #    if (!('dimnames' %in% names(x))) {
  #      x[['dimnames']] <- dimnames
  #    } else {
  #      dimnames2 <- dimnames
  #      dimnames2[names(x[['dimnames']])] <- x[['dimnames']]
  #      x[['dimnames']] <- dimnames2
  #    }
  #    i <- 1
  #    while (i <= length(extra_vars_with_default)) {
  #      if (!(extra_vars_with_default[i] %in% names(x))) {
  #        x[[extra_vars_with_default[i]]] <- replace_values[[extra_vars_with_default_ind[i]]]
  #      }
  #      i <- i + 1
  #    }
  #    x
  #  })
  #}
  #single_dataset <- (length(obs) + length(exp) == 1)
  dim_params[[found_pattern_dim]] <- dat_names

  # Check that no unrecognized variables are present in the path patterns
  # and also that no file dimensions are requested to THREDDs catalogs.
  # And in the mean time, build all the work pieces and look for the 
  # first available file of each dataset.
  array_of_files_to_load <- NULL
  indices_of_first_files_with_data <- vector('list', length(dat))
  found_file_dims <- vector('list', length(dat))
  expected_inner_dims <- vector('list', length(dat))
  for (i in 1:length(dat)) {
    dat_selectors <- dim_params
    dim_vars <- paste0('$', dim_names, '$')
    file_dims <- dim_names[which(sapply(dim_vars, grepl, dat[[i]][['path']], fixed = TRUE))]
    found_file_dims[[i]] <- file_dims
    file_dims <- unique(c(pattern_dims, file_dims))
    expected_inner_dims[[i]] <- dim_names[which(!(dim_names %in% file_dims))]
    replace_values <- vector('list', length = length(file_dims))
    names(replace_values) <- file_dims
    dataset_has_files <- FALSE
    # Take the first selector for all possible file dimensions
    for (file_dim in file_dims) {
      if (file_dim %in% names(var_params)) {
        print(paste0("The '", file_dim, "_var' param will be ignored since '", 
                     file_dim, "' is a file dimension (for the dataset with pattern ", 
                     dat[[i]][['path']], ")."))
      }
      if (is.character(dat_selectors[[file_dim]]) && 
          !((length(dat_selectors[[file_dim]]) == 1) && 
            (dat_selectors[[file_dim]] %in% c('all', 'first', 'last')))) {
        dat_selectors[[file_dim]] <- selector_checker(selectors = dat_selectors[[file_dim]],
                                                      return_indices = FALSE)
        replace_values[[file_dim]] <- dat_selectors[[file_dim]][1]
      }
    }
    replace_values[[found_pattern_dim]] <- dat_selectors[[found_pattern_dim]][i]
    # Now we know which dimensions whose selectors are provided non-explicitly.
    undefined_file_dims <- file_dims[which(sapply(replace_values, is.null))]
    defined_file_dims <- file_dims[which(!(file_dims %in% undefined_file_dims))]
    # Find the possible values for the selectors that are provided as
    # indices. If the requested file is on server, impossible operation.
    if (length(grep("^http", dat[[i]][['path']])) > 0) {
      if (length(undefined_file_dims) > 0) {
        stop(paste0("All selectors for the file dimensions must be character strings if requesting data to a remote server. Found invalid selectors for the file dimensions ",
                    paste(paste0("'", undefined_file_dims, "'"), collapse = ', '), "."))
      }
      dataset_has_files <- TRUE
    } else {
      # Iterate over the known dimensions to find the first existing file.
      # The path to the first existing file will be used to find the 
      # values for the non explicitly defined selectors.
      first_file <- NULL
      first_file_selectors <- NULL
      if (length(undefined_file_dims) > 0) {
        replace_values[undefined_file_dims] <- '*'
      }
      ## TODO: What if length of defined_file_dims is 0? code might crash (in practice it worked for an example case)
      files_to_check <- sapply(dat_selectors[defined_file_dims], length)
      files_to_check[found_pattern_dim] <- 1
      sub_array_of_files_to_check <- array(1:prod(files_to_check), dim = files_to_check)
      j <- 1
      while (j <= prod(files_to_check) && is.null(first_file)) {
        selector_indices <- which(sub_array_of_files_to_check == j, arr.ind = TRUE)[1, ]
        selectors <- sapply(1:length(defined_file_dims), function (x) dat_selectors[defined_file_dims][[x]][selector_indices[x]])
        selectors[[which(defined_file_dims == found_pattern_dim)]] <- dat_selectors[[found_pattern_dim]][i]
        replace_values[defined_file_dims] <- selectors
        file_path <- .ConfigReplaceVariablesInString(dat[[i]][['path']], replace_values)
        file_path <- Sys.glob(file_path)
        if (length(file_path) > 0) {
          first_file <- file_path[1]
          first_file_selectors <- selectors
        }
        j <- j + 1
      }
      # Start looking for values for the non-explicitly defined selectors.
      if (is.null(first_file)) {
        print(paste0("Warning: no found files for the datset '", dat[[i]][['name']], 
                     "'. Check its path pattern: ", dat[[i]][['path']]))
      } else {
        dataset_has_files <- TRUE
        ## TODO: Improve message here if no variable found:
        if (length(undefined_file_dims) > 0) {
          # Looking for the first values, parsed from first_file.
          first_values <- vector('list', length = length(undefined_file_dims))
          names(first_values) <- undefined_file_dims
          found_values <- 0
          stop <- FALSE
          try_dim <- 1
          last_success <- 1
          while ((found_values < length(undefined_file_dims)) && !stop) {
            u_file_dim <- undefined_file_dims[try_dim]
            if (is.null(first_values[[u_file_dim]])) {
              path_with_globs_and_tag <- .ConfigReplaceVariablesInString(dat[[i]][['path']], 
                                           replace_values[-which(file_dims == u_file_dim)],
                                           allow_undefined_key_vars = TRUE)
              found_value <- .FindTagValue(path_with_globs_and_tag, 
                                           first_file, u_file_dim)
              if (!is.null(found_value)) {
                found_values <- found_values + 1
                last_success <- try_dim
                first_values[[u_file_dim]] <- found_value
                replace_values[[u_file_dim]] <- found_value
              }
            }
            try_dim <- (try_dim %% length(undefined_file_dims)) + 1
            if (try_dim == last_success) {
              stop <- TRUE
            }
          }
          if (found_values < length(undefined_file_dims)) {
            stop(paste0("Path pattern of dataset '", dat[[i]][['name']], "' is too complex. Could not automatically detect values for all non-explicitly defined indices. Check its pattern: ", 
                        dat[[i]][['path']]))
          }
          ## TODO: Replace ReplaceGlobExpressions by looped call to FindTagValue? As done above
          ##       Maybe it can solve more cases actually. I got warnings in ReplGlobExp with a typical
          ##       cmor case, requesting all members and chunks for fixed var and sdate. Not fixing
          ##       sdate raised 'too complex' error.
          # Replace shell globs in path pattern and keep the file_dims as tags
          dat[[i]][['path']] <- .ReplaceGlobExpressions(dat[[i]][['path']], first_file, replace_values, 
                                                        file_dims, dat[[i]][['name']], FALSE)
          # Now time to look for the available values for the non 
          # explicitly defined selectors.
          for (u_file_dim in undefined_file_dims) {
            parsed_values <- c()
            replace_values[undefined_file_dims] <- first_values
            replace_values[[u_file_dim]] <- '*'
            path_with_globs <- .ConfigReplaceVariablesInString(dat[[i]][['path']], replace_values)
            found_files <- Sys.glob(path_with_globs)
            if (length(found_files) == 0) {
              stop("Error 3.")
            }
            for (found_file in found_files) {
              path_with_globs_and_tag <- .ConfigReplaceVariablesInString(dat[[i]][['path']],
                                           replace_values[-which(file_dims == u_file_dim)],
                                           allow_undefined_key_vars = TRUE)
              parsed_values <- c(parsed_values, 
                                 .FindTagValue(path_with_globs_and_tag, found_file, 
                                               u_file_dim))
            }
            dat_selectors[[u_file_dim]] <- selector_checker(selectors = dat_selectors[[u_file_dim]],
                                                            var = parsed_values,
                                                            return_indices = FALSE)
          }
        } else {
          dat[[i]][['path']] <- .ReplaceGlobExpressions(dat[[i]][['path']], first_file, replace_values, 
                                                        defined_file_dims, dat[[i]][['name']], FALSE)
        }
      }
    }
    if (dataset_has_files) {
      # Now fetch for the first available file
      replace_values <- vector('list', length = length(file_dims))
      names(replace_values) <- file_dims
      files_to_load <- sapply(dat_selectors[file_dims], length)
      files_to_load[found_pattern_dim] <- 1
      sub_array_of_files_to_load <- array(1:prod(files_to_load), 
                                          dim = files_to_load,
                                          dimnames = file_dims)
      j <- 1
      while (j <= prod(files_to_load)) {
        selector_indices <- which(sub_array_of_files_to_load == j, arr.ind = TRUE)[1, ]
        names(selector_indices) <- file_dims
        selectors <- sapply(1:length(file_dims), function (x) dat_selectors[file_dims][[x]][selector_indices[x]])
        selectors[[which(file_dims == found_pattern_dim)]] <- dat_selectors[[found_pattern_dim]][i]
        replace_values[file_dims] <- selectors
        file_path <- .ConfigReplaceVariablesInString(dat[[i]][['path']], replace_values)
        sub_array_of_files_to_load[j] <- file_path
        if (is.null(indices_of_first_files_with_data[[i]])) {
          if (!(length(grep("^http", file_path)) > 0)) {
            if (!file.exists(file_path)) {
              file_path <- NULL
            }
          }
          if (!is.null(file_path)) {
            test_file <- NULL
            ## TODO: suppress error messages
            test_file <- file_opener(file_path)
            if (!is.null(test_file)) {
              selector_indices[which(file_dims == found_pattern_dim)] <- i
              indices_of_first_files_with_data[[i]] <- selector_indices
              file_closer(test_file)
            }
          }
        }
        j <- j + 1
      }
      ## TODO: Put names in arrays and extend arrays as needed progressively
      if (is.null(array_of_files_to_load)) {
        array_of_files_to_load <- sub_array_of_files_to_load
      } else {
        array_of_files_to_load <- abind(array_of_files_to_load, 
                                        sub_array_of_files_to_load,
                                        along = which(file_dims == found_pattern_dim))
      }
    }
    dat[[i]][['selectors']] <- dat_selectors
  }
  if (is.null(indices_of_first_files_with_data)) {
    stop("No data files found for any of the specified datasets.")
  }

  ## TODO: To be run in parallel
  # Now time to work out the inner file dimensions
  for (i in 1:length(dat)) {
    indices <- indices_of_first_files_with_data[[i]]
    if (!is.null(indices)) {
      # First take all the variables requested.
      # For each of the dimensions across files
      # TODO: dimensions_across_files is now inverted
      # TODO: Avoid looking for the variable if the selectors are
      #       explicitly provided and the variable is not requested in 
      #       return_vars.
      picked_vars <- NULL
      for (file_dim in names(inner_dims_across_files)) {
        vars_to_pick <- NULL
        # Find the names of the requested variables that go over the 
        # dimension in var_params.
        for (inner_dim in inner_dims_across_files[[file_dim]]) {
          detected_var <- grep(inner_dim, names(var_params))[1]
          if (length(detected_var) > 0) {
            vars_to_pick <- c(vars_to_pick, var_params[[detected_var]])
          }
          # And those in return_vars.
          ## TODO: Move outisde the loops. It has to be recomputed since
          ##       return_vars may be defined over more than 1 dimension.
          #tmp_rv <- return_vars
          #return_dim_vars <- which(unlist(return_vars) %in% names(var_params))
          #if (length(return_dim_vars) > 0) {
          #  tmp_rv <- tmp_rv[-return_dim_vars]
          #}
          #vars_to_pick <- c(vars_to_pick, unlist(tmp_rv[which(names(tmp_rv) == inner_dim)], use.names = FALSE))
        }
        if (length(vars_to_pick) > 0) {
          if (indices[which(found_file_dims[[i]] == file_dim)] != 1) {
            stop(paste0("Some variables requested for the dataset '", 
                        dat[[i]][['name']], "' go along the '", file_dim, 
                        "' dimension, which goes across multiple files. ",
                        "However could not find all the files for this ",
                        "dimension. Either generate these files or ",
                        "request only for existing indices."))
          }
          picked_vars <- vector('list', length = length(vars_to_pick))
          names(picked_vars) <- vars_to_pick
          for (j in 1:(dim(array_of_files_to_load)[file_dim])) {
            indices[file_dim] <- j
            file_path <- do.call("[", c(list(array_of_files_to_load), as.list(indices)))
            file_object <- file_opener(file_path)
            if (is.null(file_object)) {
              stop(paste0("File not found for one of the indices of the dimension '", 
                          file_dim, "'. Can not read the variable(s) ", 
                          paste(paste0("'", vars_to_pick, "'") , 
                                collapse = ', '), ". ", file_path))
            }
            names(picked_vars) <- vars_to_pick
            for (var_to_pick in vars_to_pick) {
              found_values <- file_var_reader(file_path, file_object, var_to_pick)
              if (is.null(found_values)) {
                stop(paste0("Can not read the variable '", var_to_pick, 
                            "' from the file ", file_path))
              }
              picked_vars[[var_to_pick]] <- c(picked_vars[[var_to_pick]], found_values)
            }
          }
        }
      }
      for (inner_dim in expected_inner_dims) {
        picked_var <- NULL
        var_with_selectors <- var_params[[inner_dim]]
        if (!is.null(var_with_selectors)) {
          picked_var <- picked_vars[[var_with_selectors]]
        }
        dat[[i]][['selectors']][[inner_dim]] <- selector_checker(dat[[i]][['selectors']][[inner_dim]],
                                                                 var = picked_var)
      }
    }
  }
  ## TODO: Crop the found variables? Only those requested in return_vars?

  ## We know now which is the first available file. We can work out the
  ## sizes of the dimensions inside the file.
  #first_file <- file_opener(first_file_with_data)
  #first_array <- file_reader(first_file, )
  #
  ## Check all the indices provided for each dimension inside files.
  #for (dim_name in dim_names) {
  #  if (!(dim_name %in% file_dims)) {
  #    checker <- check_params[[dim_name]]
  #    if (is.null(checker)) {
  #      checker <- selector_checker
  #    }
  #    dat_selectors[[dim_name]] <- checker(dat_selectors[[dim_name]])
  #  }
  #}

  print(dat)
  #list(dim_params, dim_names, var_params, check_params, array_of_files_to_load)
}
